// Copyright 2018 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

#include "gn/test_with_scheduler.h"

TestWithScheduler::TestWithScheduler() = default;
TestWithScheduler::~TestWithScheduler() = default;
