// Copyright 2014 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

#ifndef TOOLS_GN_EXAMPLE_HELLO_STATIC_H_
#define TOOLS_GN_EXAMPLE_HELLO_STATIC_H_

const char* GetStaticText();

#endif  // TOOLS_GN_EXAMPLE_HELLO_STATIC_H_
