// Copyright 2019 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

#import <Foundation/Foundation.h>

@interface Greetings : NSObject

+ (NSString*)greet;

+ (NSString*)greetWithName:(NSString*)name;

@end
