// Copyright 2018 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

#ifndef UTIL_EXE_PATH_H_
#define UTIL_EXE_PATH_H_

#include "base/files/file_path.h"

base::FilePath GetExePath();

#endif  // UTIL_EXE_PATH_H_
